/******************************************************************************
 *
 * MotionViewer
 * Copyright (C) 2007 Sangyoon Lee, Electronic Visualization Laboratory, 
 * University of Illinois at Chicago
 *
 * This software is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this software; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Questions or comments about MotionViewer should be directed to 
 * cavern@evl.uic.edu
 *
 * Original source code from Ogre File Dialog Example
 * Refer to http://www.ogre3d.org/wiki/index.php/CEGUICommonFileDialog
 * Minor modification for mac porting
 *****************************************************************************/

#ifndef _CEGUI_HASHMAP_TEMPLATES_H_
#define _CEGUI_HASHMAP_TEMPLATES_H_

#include <CEGUI.h>
#include <ogre.h>

#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
#include <hash_map>
#else
#include <ext/hash_map>
#include "OgreString.h"
#endif

#if defined(_MSC_VER)
#	pragma warning(push)
#	pragma warning(disable : 4786)
#endif

using namespace std;
#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32
using namespace stdext;
#endif

/*
#if OGRE_PLATFORM == OGRE_PLATFORM_APPLE
namespace __gnu_cxx
{
	template <>
	struct hash<std::string>
	{
		size_t operator() (const std::string& x) const
		{
			return hash<const char*>() (x.c_str());
		};
}
#endif
*/	
namespace CEGUI
{
	/*********************************************************************************************
	Method      : getEntryFromHashMap
	Description : The following template methods provide generic methods, such as adding,
	              searching and deleting entries to/from a HashMap. The HashMap is a hash_map
	              with attributes of type Z as an identifying key and attributes of class T as
				  the object to be stored.
	              This template method is used to search an object in the HashMap, using the id
				  as search key.
	Parameters  : NA
	Returnvalue : NA
	**********************************************************************************************/
#if OGRE_PLATFORM == OGRE_PLATFORM_APPLE
	template<typename Z, class T>
	T getEntryFromHashMap (Z id, __gnu_cxx::hash_map<Z, T>* hashMap)
#else
	template<typename Z, class T>
	T getEntryFromHashMap (Z id, hash_map<Z, T>* hashMap)
#endif
	{
		// Search the entry in the hashMap
		if (hashMap)
		{
#if OGRE_PLATFORM == OGRE_PLATFORM_APPLE
			typename __gnu_cxx::hash_map<Z, T>::iterator hashMapIterator;
#else
			hash_map<Z, T>::iterator hashMapIterator;
#endif
			hashMapIterator = hashMap->find (id);
			if (hashMapIterator != hashMap->end())
			{
				// Found it
				return hashMapIterator->second;
			}
		}

		return NULL;
	};

	/*********************************************************************************************
	Method      : addEntryToHashMap
	Description : This template method is used to add an entry to the hashMap. If an entry
	              with the same key is already present, the stored object is deleted!
	Parameters  : NA
	Returnvalue : NA
	**********************************************************************************************/
#if OGRE_PLATFORM == OGRE_PLATFORM_APPLE
	template<typename Z, class T>
	void addEntryToHashMap (Z id, T object, __gnu_cxx::hash_map<Z, T>* hashMap)	
#else
	template<typename Z, class T>
	void addEntryToHashMap (Z id, T object, hash_map<Z, T>* hashMap)
#endif	
	{
		if (hashMap)
		{
			T existingObject = getEntryFromHashMap (id, hashMap);
			if (existingObject == NULL)
			{
				// Not found; insert it
				hashMap->insert (make_pair(id, object));
			}
			else
			{
				if (object != existingObject)
				{
					// The key exists, but the object differs
					hashMap->erase (id); // Remove the entry from the hash_map
					delete existingObject; // Delete the object
					hashMap->insert (make_pair(id, object));
				}
			}
		}
	};

	/*********************************************************************************************
	Method      : eraseEntryFromHashMap
	Description : This template method is used to remove one entry from the hashMap. The entry itself
	              will not be deleted!
	Parameters  : NA
	Returnvalue : NA
	**********************************************************************************************/
#if OGRE_PLATFORM == OGRE_PLATFORM_APPLE
	template<typename Z, class T>
	void eraseEntryFromHashMap (Z id, __gnu_cxx::hash_map<Z, T>* hashMap)	
#else
	template<typename Z, class T>
	void eraseEntryFromHashMap (Z id, hash_map<Z, T>* hashMap)
#endif
	{
		// Remove one entry; it wil not be deleted
		if (hashMap)
		{
#if OGRE_PLATFORM == OGRE_PLATFORM_APPLE
			typename __gnu_cxx::hash_map<Z, T>::iterator hashMapIterator;
#else
			hash_map<Z, T>::iterator hashMapIterator;
#endif
			hashMapIterator = hashMap->find (id);
			if (hashMapIterator != hashMap->end())
			{
				// Found it
				hashMap->erase (hashMapIterator);
			}
		}
	};

	/*********************************************************************************************
	Method      : deleteEntryFromHashMap
	Description : This template method is used to remove one entry from the hashMap and delete it.
	Parameters  : NA
	Returnvalue : NA
	**********************************************************************************************/
#if OGRE_PLATFORM == OGRE_PLATFORM_APPLE
	template<typename Z, class T>
	void deleteEntryFromHashMap (Z id, __gnu_cxx::hash_map<Z, T>* hashMap)	
#else
	template<typename Z, class T>
	void deleteEntryFromHashMap (Z id, hash_map<Z, T>* hashMap)
#endif
	{
		// Remove one entry; it wil not be deleted
		if (hashMap)
		{
#if OGRE_PLATFORM == OGRE_PLATFORM_APPLE
			typename __gnu_cxx::hash_map<Z, T>::iterator hashMapIterator;
#else
			hash_map<Z, T>::iterator hashMapIterator;
#endif
			hashMapIterator = hashMap->find (id);
			if (hashMapIterator != hashMap->end())
			{
				// Found it
				hashMap->erase (hashMapIterator);
				delete hashMapIterator->second;
			}
		}
	};

	/*********************************************************************************************
	Method      : deleteAllEntriesFromHashMap
	Description : This template method is used to delete all entries from the hashMap.
	Parameters  : NA
	Returnvalue : NA
	**********************************************************************************************/
#if OGRE_PLATFORM == OGRE_PLATFORM_APPLE
	template<typename Z, class T>
	void deleteAllEntriesFromHashMap (__gnu_cxx::hash_map<Z, T>* hashMap)	
#else
	template<typename Z, class T>
	void deleteAllEntriesFromHashMap (hash_map<Z, T>* hashMap)
#endif
	{
		// Delete all hashMap entries
		if (hashMap)
		{
#if OGRE_PLATFORM == OGRE_PLATFORM_APPLE
			typename __gnu_cxx::hash_map<Z, T>::iterator hashMapIterator;
#else
			hash_map<Z, T>::iterator hashMapIterator;
#endif
			for (hashMapIterator = hashMap->begin(); hashMapIterator != hashMap->end(); hashMapIterator++)
			{
				if (hashMapIterator->second)
					delete hashMapIterator->second;
			}
			hashMap->clear();
		}
	};
};
#endif

